/* Copyright 2007-2015 QReal Research Group
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. */

#pragma once

#include <QtWidgets/QGraphicsItem>
#include <QtWidgets/QGraphicsView>
#include <QtXml/QDomDocument>
#include <QGraphicsSceneHoverEvent>

#include "qrutils/utilsDeclSpec.h"

const int drift = 15;
const int resizeDrift = 10;
const int scalingDrift = 10;
const int scalingRect = 6;

namespace graphicsUtils {

class QRUTILS_EXPORT AbstractItem : public QGraphicsObject
{
	Q_OBJECT

public:
	enum DragState {
		None
		, TopLeft
		, TopRight
		, BottomLeft
		, BottomRight
		, Ctrl
	};

	enum ZValue {
		Background
		, Region
		, Picture
		, Shape
		, Marker
		, Wall
		, Moveable
		, Robot
	};

	explicit AbstractItem(QGraphicsItem *parent = nullptr);

	virtual QRectF boundingRect() const override = 0;
	virtual QRectF realBoundingRect() const;
	virtual QRectF calcNecessaryBoundingRect() const;
	virtual QPainterPath realShape() const;
	virtual QPainterPath resizeArea() const;
	virtual void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget = nullptr) override;
	virtual void drawItem(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget = nullptr) = 0;
	virtual void drawExtractionForItem(QPainter *painter);
	virtual void drawFieldForResizeItem(QPainter *painter);
	virtual void setPenBrushForExtraction(QPainter *painter, const QStyleOptionGraphicsItem *option);
	virtual void setPenBrushDriftRect(QPainter *painter);
	virtual QDomElement serialize(QDomElement &parent) const;
	virtual void deserialize(const QDomElement &element);

	static QStringList getPenStyleList();
	static QStringList getBrushStyleList();

	QPen pen() const;
	QBrush brush() const;
	void setBrush(const QBrush &brush);
	void setPen(const QPen &pen);

	virtual void setPenStyle(const QString &text);
	virtual void setPenWidth(int width);
	virtual void setPenColor(const QString &text);
	virtual void setBrushStyle(const QString &text);
	virtual void setBrushColor(const QString &text);
	virtual void setBrush(const QString &brushStyle, const QString &brushColor);
	virtual void setPen(const QString &penStyle, int width, const QString &penColor);
	virtual void setPenBrush(const QString &penStyle, int width, const QString &penColor, const QString &brushStyle
			, const QString &brushColor);

	/// Returns a x-coordinate of the first item`s end (it may be one corner of the rectangle or the begin of the line).
	qreal x1() const;
	/// Returns a y-coordinate of the first item`s end (it may be one corner of the rectangle or the begin of the line).
	qreal x2() const;
	/// Returns a x-coordinate of the second item`s end (it may be one corner of the rectangle or the end of the line).
	qreal y1() const;
	/// Returns a y-coordinate of the second item`s end (it may be one corner of the rectangle or the end of the line).
	qreal y2() const;
	/// Sets an x-coordinate of the first item`s end (it may be one corner of the rectangle or the begin of the line).
	void setX1(qreal x1);
	/// Sets an y-coordinate of the first item`s end (it may be one corner of the rectangle or the begin of the line).
	void setY1(qreal y1);
	/// Sets an x-coordinate of the second item`s end (it may be one corner of the rectangle or the end of the line).
	void setX2(qreal x2);
	/// Sets an y-coordinate of the second item`s end (it may be one corner of the rectangle or the end of the line).
	void setY2(qreal y2);

	void setCoordinates(const QRectF &pos);

	virtual void reshapeRectWithShift();
	virtual void changeDragState(qreal x, qreal y);
	DragState dragState() const;
	void setDragState(DragState dragState);

	virtual void calcResizeItem(QGraphicsSceneMouseEvent *event);
	virtual void resizeItem(QGraphicsSceneMouseEvent *event);
	void reverseOldResizingItem(const QPointF &begin, const QPointF &end);

	virtual void restorePos();
	virtual void savePos();

	//for save to xml
	virtual void setXandY(QDomElement& dom, const QRectF &rect);
	QDomElement setPenBrushToDoc(QDomDocument &document, const QString &domName) const;
	QDomElement setPenBrushToElement(QDomElement &target, const QString &domName) const;
	virtual QRectF sceneBoundingRectCoord(const QPoint &topLeftPicture);
	void readPenBrush(const QDomElement &docItem);

	void mouseMoveEvent(QGraphicsSceneMouseEvent *event) override;

	/// Returns a unique identifier of an item if it was assigned in the model.
	QString id() const;

	/// Sets a unique identifier of an item.
	void setId(const QString &id);

	virtual void setEditable(bool editable);
	bool editable() const;
	bool isHovered() const;

	void mouseReleaseEvent(QGraphicsSceneMouseEvent *event) override;

	Qt::CursorShape getResizeCursor();

	QPen getStrokePen();

signals:
	/// Emitted when QGraphicsItem has repositioned.
	void positionChanged(const QPointF &pos);

	/// Emitted when the x-coorinate of the first item`s end modified for some reason.
	void x1Changed(qreal x1);

	/// Emitted when the y-coorinate of the first item`s end modified for some reason.
	void y1Changed(qreal y1);

	/// Emitted when the x-coorinate of the second item`s end modified for some reason.
	void x2Changed(qreal x2);

	/// Emitted when the y-coorinate of the second item`s end modified for some reason.
	void y2Changed(qreal y2);

	/// Emitted when item`s pen changed somehow.
	void penChanged(const QPen &pen);

	/// Emitted when item`s brush changed somehow.
	void brushChanged(const QBrush &brush);

	/// Emitted when user clicks with right mouse button on this item and selects to remove it.
	void deletedWithContextMenu();

	/// Emitted when user presses left mouse button on this item.
	void mouseInteractionStarted();

	/// Emitted when user releases left mouse button on this item.
	void mouseInteractionStopped();

protected:
	explicit AbstractItem(Qt::CursorShape resizeCursor, QGraphicsItem *parent = nullptr);

	void contextMenuEvent(QGraphicsSceneContextMenuEvent *event) override;
	QVariant itemChange(GraphicsItemChange change, const QVariant &value) override;

	void mousePressEvent(QGraphicsSceneMouseEvent *event) override;
	void hoverEnterEvent(QGraphicsSceneHoverEvent *event) override;
	void hoverMoveEvent(QGraphicsSceneHoverEvent *event) override;
	void hoverLeaveEvent(QGraphicsSceneHoverEvent *event) override;

	void copyTo(AbstractItem * const other) const;
	qreal alignedCoordinate(qreal coord, const int indexGrid) const;
	void setXYWithDragState(const QPointF pos);

private:
	DragState mDragState {None};
	QPen mPen;
	QBrush mBrush;
	qreal mX1 {0};
	qreal mY1 {0};
	qreal mX2 {0};
	qreal mY2 {0};
	QString mId;
	bool mEditable {true};
	bool mHovered {false};

	QPointF mOldPos;
	qreal mOldX1 {};
	qreal mOldY1 {};
	qreal mOldX2 {};
	qreal mOldY2 {};

	QPen mStrokePen {Qt::green};
	Qt::CursorShape mResizeCursor {Qt::SizeAllCursor};
	const Qt::CursorShape mHoverCursor {Qt::PointingHandCursor};
};

}
